import numpy as np


# TODO: Implement this function
'''
Returns (IsValid, Message)
'''
def validate_replicatesIDs(replicatesIDs, X):
    return True, None


# TODO: Implement this function
'''
Returns (IsValid, Message)
'''
def validate_normalisation(normalisationCodes, X):
    return True, None